/*******************************************************************************
 *                                                                             *
 *                          CIS600 Homework #5                                 *
 *                                                                             *
 *                            SEUNGJO BAE                                      *
 *                                                                             *
 *   - Multiplication of two N x N matrices (of type double).                  *
 *                                                                             *
 *   - Version 2.0(11/17/93)                                                   *
 *                                                                             *
 *   - Communication:                                                          *
 *                                                                             *
 *       [ No-Pipelining ]                                                     *
 *          1-to-all BC : Blocking Send & Blocking Receive                     *
 *          Shift       : Collective Send_and_Receive                          *
 *                                                                             *
 *       [ Pipelining ]                                                        *
 *          1-to-all BC : Asynchronous Send & Blocking Receive                 *
 *          Shift       : Non-blocking Send and Blocking Receive               *
 *                                                                             *
 *******************************************************************************/


#include <cm/cmmd.h>
#include <stdio.h>
#include <stdlib.h>

#define MAXINT  2147483647.              /* for random number generator */
#define range   10.                      /*         "         "         */

#define Nmax    2048                     /* maximum matrix size */

#define m(r, c)  *(M  + r*Nlocal + c)    /* for dynamic 2-dim. array    */
#define m1(r, c) *(M1 + r*Nlocal + c)    /*         "         "         */
#define m2(r, c) *(M2 + r*Nlocal + c)    /*         "         "         */
#define m3(r, c) *(M3 + r*Nlocal + c)    /*         "         "         */

#define Row       0
#define Column    1

typedef enum {false, true} Boolean;


/*******************************************************************************
 *                                                                             *
 *   Global variables                                                          *
 *                                                                             *
 *******************************************************************************/

double *A, *B, *C, *X;       /* N x N matrix */

int N,        /* matrix size of A, B, C such that N is a power of 2 */
    Nlocal;   /* local sub matrix size */
    
int ori_nproc,              /* original number of processors */
    nproc,                  /* number of processors */
    pnum,                   /* processor number */
    mesh_size,              /* size of 2-D mesh in each direction */
                            /* mesh_size is a power of 2 */
    coord[2],               /* logical coordinate of a processor on 2-D mesh */
    row_num, col_num;       /* coord[0] = row_num, coord[1] = col_num */

int buf_size,               /* communication buffer size */
    up_chan, down_chan,     /* 4 neighboring processors for each processor */
    right_chan, left_chan;
    
int seed;                   /* seed for random number generator */



/*******************************************************************************
 *  Boolean Function : Blog2(n)                                                *
 *                                                                             *
 *   - Input  = positive integer                                               *
 *   - Return = true if log2(n) is integer                                     *
 *******************************************************************************/

Boolean Blog2(n)
     int n;
{
  int x = n, i = 0;

  if ( n < 1 ) return false;

  while ( n > 1 ) {
     i += 1;
     n /= 2;
   };

  return ( x == ( 1 << i ) );
};


/*******************************************************************************
 *  Integer Function : Ilog2(n)                                                *
 *                                                                             *
 *   - Input  = positive integer                                               *
 *   - Return = log2(n) if log2(n) is integer                                  *
 *              otherwise return -1                                            *
 *******************************************************************************/

int Ilog2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1<<i != n )
    return -1;
  else
    return i;
}


/*******************************************************************************
 *  Integer Function : node_to_grid(num, coord)                                *
 *                                                                             *
 *   - Input  = physical processor-number                                      *
 *   - Output = logical grid coordinate on 2-D mesh                            *
 *******************************************************************************/

int node_to_grid(num, coord)
   int num;
   int coord[2];
   
{
   if ( num < 0 || num > nproc )
      return (-1);

   coord[0] = num / mesh_size;   /* row index    */
   coord[1] = num % mesh_size;   /* column index */
   return (0);
}


/*******************************************************************************
 *  Integer Function : grid_to_node(coord)                                     *
 *                                                                             *
 *   - Input  = logical grid coordinate on 2-D mesh                            *
 *   - Return = physical processor-number                                      *
 *******************************************************************************/

int grid_to_node(coord)
   int coord[2];
{
   int num;
 
   num = coord[0] * mesh_size + coord[1];

   if ( num < 0 || num > nproc )
      return (-1);
   else
      return num;
}


/*******************************************************************************
 *  Integer Function : gridchan(num, dim, dir)                                 *
 *                                                                             *
 *   - Input  = processor-number, dimension, direction.                        *
 *   - Return = corresponding processor number according to the input.         *
 *              -1 if input is improper.                                       *
 *******************************************************************************/

int gridchan(num, dim, dir)
   int num, dim, dir;
{
   int coord[2];
   int proc;

   if (dim != 0 && dim !=1)
      return(-1);

   if ( node_to_grid(num,coord) == -1 )
      return(-1);

   coord[dim] = ( coord[dim] + dir ) % mesh_size;  

   if ( coord[dim] < 0 )
       coord[dim] += mesh_size;  

   return( grid_to_node(coord) );    

}


/*******************************************************************************
 *  double Function : drandom()                                                *
 *                                                                             *
 *   - Return = random number r (of type double) where  0 <= r <= range        *
 *******************************************************************************/

double drandom()
{
   double r;

   r = (double)random();
   return ( (r / MAXINT) * range);

};


/******************************************************************************
 *  Function : printing()                                                     *
 *                                                                            *
 *    - Task : Print out the matrix                                           *
 ******************************************************************************/

void printing(M)
     double *M;
{
  int i, j,          /* global index */
      r, c,          /* local index */
      grid[2],       /* index for submatrix */
      pn, s;
  double x;

  CMMD_sync_with_nodes();
  CMMD_fset_io_mode(stdout, CMMD_independent);

  s = sizeof(double);
  for (i=0; i<N; ++i) {
    grid[0] = i / Nlocal;
    r       = i % Nlocal;

    for (j=0; j<N; ++j) {
      CMMD_sync_with_nodes();
      grid[1] = j / Nlocal;
      c       = j % Nlocal;
      pn = grid_to_node(grid);
      x = m(r, c);

      if ( pnum == 0 ) {
	if ( pn == 0 ) 
	  printf("%3.0f", x);
	else {
          CMMD_receive_block(pn, CMMD_ANY_TAG, &x, s);
          printf("%3.0f", x);
	}
      }	  
      else if ( pnum == pn ) 
	CMMD_send_block(0, CMMD_DEFAULT_TAG, &x, s); 
    }
    if ( pnum == 0 )  printf("\n");

  }

   CMMD_sync_with_nodes();
   CMMD_fset_io_mode(stdout,CMMD_sync_bc);
   printf("\n\n");

};   


/*******************************************************************************
 *  void Function : initialize()                                               *
 *                                                                             *
 *   - Task :                                                                  *
 *       (1) Set up CMMD system variables                                      *
 *       (2) initialize global variable                                        *
 *       (3) generate random matrix A, B, and partition                        *
 *******************************************************************************/

void initialize()
{
  int i;
  void rand_mat();

  CMMD_disable_safety();
  CMMD_fset_io_mode(stdin, CMMD_sync_bc);
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  CMMD_fset_io_mode(stderr, CMMD_independent);

  buf_size = Nlocal * Nlocal * sizeof(double);

  ori_nproc = CMMD_partition_size();

  pnum = CMMD_self_address(); 

  nproc = 32;

  i = CMMD_reset_partition_size(nproc);
  if (i != nproc) 
    CMMD_error("\nERROR: CMMD Partition failed!\n");

}



/*******************************************************************************
 *  void Function: rand_mat(M)                                                 *
 *                                                                             *
 *   - Task :  Generate random matrix M                                        *
 *******************************************************************************/

void rand_mat(M)
     double *M;
{
  int i, j,          /* global index */
      r, c,          /* local index */
      rsub, csub;    /* index for submatrix */

  for (i=0; i<N; ++i) {
    rsub = i / Nlocal;
    r    = i % Nlocal;
    for (j=0; j<N; ++j) {
      csub = j / Nlocal;
      c    = j % Nlocal;
      if ( rsub==row_num && csub==col_num)
	m(r,c) = drandom();
    }
  }
}



/*******************************************************************************
 *  void Function : row_bc(mat, buf, step)                                     *
 *                                                                             *
 *   - Input = matrix, communication buffer, current step                      *
 *   - Task  = 1-to-all broadcast through row(cut-through routing)             *
 *             (Kurmar's book pp75-76)                                         *
 *   - Note  = No pipelining                                                   *
 *******************************************************************************/

void row_bc(mat, buf, step)  
     double *mat, *buf;
     int step;
{
  int i, j;
  int first,      /* column number of source for 1-to-all bc */
      dist;       /* distancd from source to destination */
  Boolean source = false, 
          dest = false;     /* mask for communication */
  int index;      /* new column index for 1-to-all bc:                   */ 
                  /* col_num(0...mesh_size-1) ==> index(0...mesh_size-1) */
                  /* source node has index 0                             */

  first = (row_num + step) % mesh_size;
  index  = col_num - first;
  if ( index < 0 ) 
    index += mesh_size;

  if (!index) {
    source =true;
    for (j=0; j<Nlocal*Nlocal; ++j)    /* prepare communication buffer */
      buf[j] = mat[j];
  }

  dist = mesh_size >> 1;

  for (i = 1; i <= Ilog2(mesh_size); ++i) {  /* dist = mesh_size / 2^i */
    dest = !(index % dist) && !source; 
    
    if (source) 
      CMMD_send_block( gridchan(pnum, Column, dist), CMMD_DEFAULT_TAG, 
                       buf, buf_size);
    
    if (dest) {
      CMMD_receive_block( gridchan(pnum, Column, -dist), CMMD_DEFAULT_TAG, 
                       buf, buf_size);
      source = true;    /* this node will be a source in the next iteration */
    };

    dist >>= 1;         
  }
}



/*******************************************************************************
 *  void Function : row_bc_pipe(mat, buf, step)                                *
 *                                                                             *
 *   - Input = matrix, communication buffer, current step                      *
 *   - Task  = 1-to-all broadcast through row by using pipelining              *
 *   - Note  = pipelining                                                      *
 *******************************************************************************/
  
void row_bc_pipe(mat, buf, step)  
     double *mat, *buf;
     int step;
{
  
  int first,   /* column number of source for 1-to-all bc */
      last,    /* column number of left neighbor of source */
      i;


  first = (row_num + step) % mesh_size;
  last  = (!first) ? mesh_size - 1 : first - 1;

  if ( col_num == first ) {
    for (i=0; i<Nlocal*Nlocal; ++i)    /* prepare communication buffer */
      buf[i] = mat[i];
    CMMD_free_mcb(
     CMMD_send_async( right_chan, CMMD_DEFAULT_TAG, buf, buf_size, NULL, NULL));
  }
  else if ( col_num == last )
    CMMD_receive_block( left_chan, CMMD_DEFAULT_TAG, buf, buf_size);
  else {
    CMMD_receive_block( left_chan, CMMD_DEFAULT_TAG, buf, buf_size);
    CMMD_free_mcb(
     CMMD_send_async( right_chan, CMMD_DEFAULT_TAG, buf, buf_size, NULL, NULL));
  }

}


/*******************************************************************************
 *  void Function : mat_mult(M1, M2, M3)                                       *
 *                                                                             *
 *   - Input  = matrix M1 and M2                                               *
 *   - Output = matrix M3 such that M3 = M1 * M2                               *
 *******************************************************************************/

void mat_mult(M1, M2, M3)
     double *M1, *M2, *M3;
{
  double tmp;
  int i, j, k;

  for (i=0; i<Nlocal; ++i)
    for (j=0; j<Nlocal; ++j) {
      tmp = m3(i,j);
      for(k=0; k<Nlocal; ++k)
	tmp += m1(i,k) * m2(k,j);
      m3(i,j) = tmp;
    }
}



/*******************************************************************************
 *  void Function : col_shift(M)                                               *
 *                                                                             *
 *   - Input = matrix M                                                        *
 *   - Task  = shift up M through column                                       *
 *******************************************************************************/

void col_shift(M)    
     int *M;
{
   CMMD_send_and_receive(  /* receive from the DOWN */
                           down_chan,
                           CMMD_DEFAULT_TAG,
                           M,
                           buf_size,
                           /* send to the UP */
                           up_chan, 
                           CMMD_DEFAULT_TAG,
                           M, 
                           buf_size);
}



/*******************************************************************************
 *  void Function : col_shift_pipe(M)                                          *
 *                                                                             *
 *   - Input = matrix M                                                        *
 *   - Task  = shift up M through column                                       *
 *******************************************************************************/

void col_shift_pipe(M)    
     int *M;
{
   CMMD_send_noblock( up_chan,   CMMD_DEFAULT_TAG, M, buf_size);

   CMMD_receive_block(down_chan, CMMD_DEFAULT_TAG, M, buf_size);
}



/*******************************************************************************
 *                                                                             * 
 *                              MAIN  PROGRAM                                  *
 *                                                                             *
 *******************************************************************************
 *                                                                             *
 *  Command line: matmul N mesh_size {seed}                                    *
 *                                                                             *
 *   -  argv[1] = Matrix size                                                  *
 *   -  argv[2] = Number of processors on 2-D mesh in each direction           *
 *   -  argv[3] = (optional) seed for a random number generator                *
 *                                                                             *
 *******************************************************************************/
  
main(argc, argv)
     int argc;
     char *argv[];
{
  void initialize();
  void row_bc();
  void row_bc_pipe();
  void mat_mult();
  void col_shift();

  int    i, step, buf[1];
  double time, max0, max1, max2, max3;


  /*---------------------  Initialize and Print out input ---------------------*/

  initialize();

/*
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n");
  printf("***********************************************\n");
  printf("                 Input Data  \n");
  printf("***********************************************\n");
  printf("   Number of processors = %4d\n", nproc);
*/

  buf[0] = pnum;

  if ( pnum == 0 ) {
    for ( i = 1 ; i < nproc ; i++ ) {
      CMMD_receive_block( i, CMMD_DEFAULT_TAG, buf, sizeof(int));

      CMMD_fset_io_mode(stdout, CMMD_sync_bc);
      printf("From %d \n", buf[0]);
/*      CMMD_sync_with_nodes();*/

    }
  }
  else {
    CMMD_send_noblock( 0,   CMMD_DEFAULT_TAG, buf, sizeof(int));
  }

  /*---------------------  Print out the timiing result   ---------------------*/
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n");
  printf("***********************************************\n");
  printf("                 Input Data  \n");
  printf("***********************************************\n");
  printf("   Number of processors = %4d\n", nproc);

}

